import torch.utils.data as datalodaer
from framework.dataloader.Transform import TRANSFORM
from abc import *


class BaseDataset(datalodaer.Dataset, metaclass=ABCMeta):
    def __init__(self, example: list, required: list, meta, config):
        """
        [configuration unsafe]
        Dataset Class works independently from App Class Configuration. So, it doesn't matter if the configuration is unsafe.

        :param example: The example must have information that can load all sample data. e.g., path information
        :param required: required_input
        :param meta: meta defines TensorTypes of the required data. When the class is inherited, the type for internal information must be specified.
        :param config: The dataset class is independent of the App. Therefore, it does not matter if it is corrected.
        """
        self._examples: list = example
        self._required: list = required
        self._meta = meta
        self.config = config
        self._transformers = []
        self.make_transforms()

    def make_transforms(self):
        """
        Create a transform method based on the transform recorded in the configuration.
        format: [TransformName:INPUT1-INPUT2-...]
        e.g., [ToTensor:IMAGE-GT]
        """
        for transforms_info in self.config.transforms:
            transforms_info = transforms_info.split(':')
            transform_name = transforms_info[0]
            required_inputs = set(self._required)
            if len(transforms_info) > 1:
                required_inputs = required_inputs.intersection(set(transforms_info[1].split('-')))
            self._transformers.append(TRANSFORM[transform_name](list(required_inputs), self._meta))

    @classmethod
    @abstractmethod
    def datasetFactory(self, config):
        raise NotImplementedError

    def __getitem__(self, item):
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError




import torch
import torch.utils.data as datalodaer
from framework.app.app import *
from framework.dataloader.Dataset import BaseDataset
from framework.utils.runtime import get_class_object_from_name
import copy

class DataLoaderController(SingletoneInstance):
    def __init__(self):
        """
        DataLoaderController: SingletoneInstance
        This class self.dataloader_config is [configuration safe]
        """
        self.dataloader_config = App.instance().config.DATA_LOADER
        self.dataloaders = {}

    def make_dataset(self, loader_name: str, dataloader_config=None):
        """
        This function injects the DataLoader class into self.dataloaders.

        :rtype: torch.utils.data.DataLoader
        :param loader_name: The name of the data loader
        :param dataloader_config: Configuration for dataloader
        :return: dataloader class object
        """
        default_loader = copy.deepcopy(self.dataloader_config.required_loader['default'])

        try:
            if dataloader_config is None:
                dataloader_config = default_loader.update(self.dataloader_config.required_loader[loader_name])
            else:
                dataloader_config = default_loader.update(dataloader_config)
            dataset_path = dataloader_config.dataset
            dataset_config = Config.from_yaml(dataset_path).DATASET

        except FileNotFoundError:
            print('File not founded in {}, mode:{}'.format(dataloader_config.dataset, loader_name))
            return

        if not dataloader_config.config.reload and loader_name in self.dataloaders.keys():
            return self.dataloaders[loader_name]

        dataset_module = dataset_config.module_name
        dataset_class = dataset_config.class_name
        dataset_module: BaseDataset = get_class_object_from_name(dataset_module, dataset_class)

        # dataloader_config
        dataset_args = {}
        dataset_args['name'] = loader_name
        if 'dataset_args' in dataloader_config.keys():
            dataset_args = dataloader_config.dataset_args
        else:
            dataset_args = dataset_config.args[loader_name] if loader_name in dataset_config.args else \
            dataset_config.args['default']

        dataset_args['required_input'] = dataloader_config[
            'required_input'] if "required_input" in dataloader_config.keys() else dataset_args['required_input']
        dataset_args['transforms'] = dataloader_config['transforms'] if "transforms" in dataloader_config.keys() else \
        dataset_args['transforms']
        dataset = dataset_module.datasetFactory(dataset_args)

        self.dataloaders[loader_name] = self.dataloaderFactory(dataset, dataloader_config)
        return self.dataloaders[loader_name]

    def dataloaderFactory(self, dataset, config):
        """
        :param dataset: Required Type: BaseDataset
        :param config: Config or Dictionary
        :return:
        """
        class CustomDataLoader(torch.utils.data.DataLoader):
            def __init__(self, dataset, config, **kwargs):
                super(CustomDataLoader, self).__init__(dataset, **kwargs)
                self.config = config
        return CustomDataLoader(dataset, config.config, **Config.extraction_dictionary(config.loader_args))



import os
import yaml
import re
import copy

loader = yaml.FullLoader
loader.add_implicit_resolver(
    u'tag:yaml.org,2002:float',
    re.compile(u'''^(?:
     [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
    |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
    |\\.[0-9_]+(?:[eE][-+][0-9]+)?
    |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
    |[-+]?\\.(?:inf|Inf|INF)
    |\\.(?:nan|NaN|NAN))$''', re.X),
    list(u'-+0123456789.'))

class Config(object):
    def __init__(self, dict_config=None):
        super().__init__()
        self.set_attribute(dict_config)

    @staticmethod
    def from_yaml(path):
        with open(path, 'r') as stream:
            return Config(yaml.load(stream, Loader=loader))

    @staticmethod
    def from_dict(dict):
        return Config(dict)

    @staticmethod
    def get_empty():
        return Config()

    def __getattr__(self, item):
        return self.__dict__[item]

    def __setattr__(self, key, value):
        self.set_attribute({key:value})

    def set_attribute(self, dict_config):
        if dict_config is None:
            return

        for key in dict_config.keys():
            if isinstance(dict_config[key], dict):
                self.__dict__[key] = Config(dict_config[key])
            else:
                self.__dict__[key] = dict_config[key]

    def get(self, key, default, possible_none=True):
        """
        same dictionary.get. this method Configruation Not safe!!!
        :param key:
        :param default:
        :param possible_none: if true, None also returned
        :return:
        """
        out = self.__dict__.get(key, default)
        if possible_none is False and out is None:
            return default
        return out

    def keys(self):
        return self.__dict__.keys()

    def __getitem__(self, key):
        return self.__dict__[key]

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __delitem__(self, key):
        del self.__dict__[key]

    def __contains__(self, key):
        return key in self.__dict__

    def __len__(self):
        return len(self.__dict__)

    def __repr__(self):
        return repr(self.__dict__)

    def copy(self):
        return Config.from_dict(self.extraction_dictionary(self))

    def update(self, dict_config):
        """
        Combines the current configuration with the parameter configuration.
        This operation performs a deepcopy on the parameter configuration.
        Therefore, it is configuration safe.

        :param dict_config:
        :return:
        """
        for key in dict_config.keys():
            if key in self.__dict__.keys():
                if isinstance(dict_config[key], Config):
                    self.__dict__[key].update(dict_config[key])
                else:
                    self.__dict__[key] = copy.deepcopy(dict_config[key])
            else:
                self.__setitem__(key, copy.deepcopy(dict_config[key]))
        return self

    @classmethod
    def extraction_dictionary(cls, config):
        out = {}
        for key in config.keys():
            if isinstance(config[key], Config):
                out[key] = cls.extraction_dictionary(config[key])
            else:
                out[key] = copy.deepcopy(config[key])
        return out

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v, memo))
        return result


from importlib import import_module

def get_instance_from_name(module_path, class_name, *args, **kwargs):
    m = import_module(module_path)
    return getattr(m, class_name)(*args, **kwargs)

def get_class_object_from_name(module_path, class_name):
    m = import_module(module_path)
    return getattr(m, class_name)



import torch
import torch.nn as nn
from abc import abstractmethod
from framework.utils.config import Config
from framework.app.app import App
import os


class LossFactory(object):
    @staticmethod
    def instantiate_loss(name, config):
        if config.loss_type == 'L1':
            return L1Loss(name, config)
        else:
            raise NotImplementedError(name)

class LossContainer(nn.Module):
    def __init__(self, config):
        super().__init__()

        self.loss_dict = nn.ModuleDict()
        for loss_name in config.keys():
            self.loss_dict[loss_name] = LossFactory.instantiate_loss(loss_name, config[loss_name])
        self.total_loss = LossEmpty('TotalLoss')
        self.device = App.instance().get_device()

    def forward(self, sample):
        total_losses = []
        for key in self.loss_dict.keys():
            inputs = []
            for name in self.loss_dict[key].required:
                inputs.append(sample[name].to(self.device))
            loss = self.loss_dict[key](*inputs)
            loss = self.loss_dict[key].weight * loss
            total_losses.append(loss)

        total_loss = sum(total_losses)
        self.total_loss(total_loss)
        return total_loss

    def save(self, info):
        state = info['state']
        contents = App.instance().current_time_format() + " {} epoch, {} step\n".format(state.epoch, state.step)
        contents += "|{}: {}\t".format('Total', str(self.total_loss))
        for key in self.loss_dict.keys():
            contents += "|{}: {}\t".format(key, str(self.loss_dict[key]))
        App.instance().smart_write(contents + '\n', info['path'], 'a+')

    def load(self):
        pass

    def get_dir(self, dir_path):
        return os.path.join(dir_path, 'loss.txt')

class BaseLoss(nn.Module):
    def __init__(self, name=None, config=None):
        super(BaseLoss, self).__init__()
        self.config = config

        self.name = name
        self.weight = config.weight
        self.required = config.inputs

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    @abstractmethod
    def update_state(self, loss):
        raise NotImplementedError

class L1Loss(BaseLoss):
    def __init__(self, name, config):
        super().__init__(name, config)
        self.loss = nn.L1Loss()

        self.register_buffer('sum', torch.zeros(1))
        self.register_buffer('count', torch.zeros(1))

    def forward(self, pred, target):
        loss = self.loss(pred, target)
        self.update_state(loss)
        return loss

    def update_state(self, loss):
        self.sum += loss.item()
        self.count += 1

    def __str__(self):
        if self.count==0:
            return "0"
        return str(self.sum.item()/self.count.item())

class LossEmpty(BaseLoss):
    def __init__(self, name, config=None):
        super().__init__(name, Config.from_dict({'weight':1, 'inputs':None}))

        self.register_buffer('sum', torch.zeros(1))
        self.register_buffer('count', torch.zeros(1))

    def forward(self, loss):
        self.update_state(loss)

    def update_state(self, loss):
        self.sum += loss.item()
        self.count += 1

    def __str__(self):
        if self.count==0:
            return "0"
        return str(self.sum.item()/self.count.item())

if __name__ == '__main__':
    config = Config.from_yaml('../../reso   urce/configs/trainer/ExampleController.yaml')
    print(config.MODEL_CONTROLLER.LOSSES)
    config.MODEL_CONTROLLER.LOSSES['image_loss_2'] = Config.from_dict({
        'inputs': ['OUTPUT','GT'], 'loss_type': 'L1', 'weight': 2
    })
    loss = LossContainer(config.MODEL_CONTROLLER.LOSSES)
    a = torch.tensor([4.0,1.0]).view(1,1,1,2).float()
    b = torch.tensor([7.0,3.0]).view(1, 1, 1, 2).float()
    sample = {'OUTPUT':a, 'GT':b}
    loss(sample)
    a = torch.tensor([4.0,1.0]).view(1,1,1,2).float()
    b = torch.tensor([7.0,4.0]).view(1, 1, 1, 2).float()
    sample = {'OUTPUT':a, 'GT':b}
    print(loss.total_loss)
    loss(sample)
    print(loss.total_loss)
    print(loss.loss_dict['image_loss_l1'])


import sys
import socket
from framework.ipc.RunningCommand import *
from framework.ipc.RunningModuleVisualizer import *
from struct import unpack
from framework.app.Exceptions import *


class RunnableModuleFactory(object):
    def __init__(self, runnable_controller, host, port):
        self.runnable_controller:RunnableModuleController = runnable_controller

        self.host = host
        self.port = port
        self.client_socket = None
        self.writer_process = None

        self.job_queue = queue.Queue()
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server_socket.bind((self.host, self.port))

        self.live = True
        self.sender_thread = None
        self.receiver_thread = None

    def update_graph(self):
        if not self.job_queue.empty():
            while True:
                opt, commands = self.job_queue.get()
                opt = opt.upper()
                self.runnable_controller.get_current_main_module().write_log('[IPC receive] {} {}'.format(opt, commands))
                if opt == 'ADD':
                    self.add_from_script(commands)
                elif opt == 'REMOVE':
                    self.remove_from_script(commands)
                elif opt == 'LOAD':
                    self.update_from_yaml(commands)
                if self.job_queue.empty(): break

    def update_from_yaml(self, command_path):
        config = Config.from_yaml(command_path)
        self.runnable_controller._runnable_define.update(config.DEFINE)

        for key in config.keys():
            if key == 'DEFINE' or not isinstance(config[key], list): continue
            for script in config.get(key, [], False):
                self.job_queue.put((key, script))

    def remove_from_script(self, commands):
        try:
            m = self.runnable_controller.get_runnable_module(commands)
            m.finish = True
            m.live = False
        except RunnableModuleDestroyException as e:
            pass

    def start(self):
        threading.Thread(target=self.server_socket_run).start()
        self.sender_thread = threading.Thread(target=self.sender_run)
        self.sender_thread.start()

    def end(self):
        self.live = False
        if self.client_socket is not None:
            self.client_socket.close()
        self.server_socket.close()

        if self.receiver_thread is not None:
            self.receiver_thread.join()
        if self.sender_thread is not None:
            self.sender_thread.join()

    def __del__(self):
        self.end()

    def server_socket_run(self):
        while self.live:
            self.client_socket = None
            self.server_socket.listen()
            try:
                self.client_socket, addr = self.server_socket.accept()
                self.receive()
            except Exception as e:
                print(e)
                self.client_socket.shutdown(socket.SHUT_WR)
                self.client_socket.close()

    def receive(self):
        while self.live:
            byte_len = self.client_socket.recv(8)
            (length,) = unpack('>Q', byte_len)
            buffer = b''
            while len(buffer) < length:
                # doing it in batches is generally better than trying
                # to do it all in one go, so I believe.
                to_read = length - len(buffer)
                buffer += self.client_socket.recv(
                    4096 if to_read > 4096 else to_read)

            # send our 0 ack
            assert len(b'\00') == 1
            self.client_socket.sendall(b'\00')

            data = buffer.decode().split(' ', maxsplit=1)
            if len(data) <= 1:
                continue
            self.job_queue.put((data[0].strip(), data[1].strip()))

    def sender_run(self):
        while self.live:
            self.sender()
            time.sleep(1)

    def sender(self):
        workspace = self.runnable_controller.config.get('save_path', '$base/ipc', possible_none=False)
        saved_path = App.instance().dir_path_parser(os.path.join(workspace, 'monitor.txt'))
        with open(saved_path, 'w') as f:
            f.write("{} {}\n".format(self.host, self.port))
            paths = DisplayRunnableModule.make_tree(self.runnable_controller, 'root')
            f.writelines([path.displayable(self.runnable_controller) + '\n' for path in paths])

    @staticmethod
    def parsing(line, separator='->', opener='(', closer=')'):
        res = []
        count = 0
        temp = ""
        for c in line:
            if c == opener: count += 1
            elif c == closer: count -= 1

            if count == 0 and c == separator[0]:
                res.append(temp)
                temp = ''
            elif count == 0 and c in separator:
                pass
            else:
                temp += c
        if len(temp) > 0: res.append(temp)
        return res

    def connect_edge(self, a, b):
        self.runnable_controller.next[a].append(b)
        self.runnable_controller.get_runnable_module(b).indegree += 1
        self.runnable_controller.previous[b].append(a)

    def create_module(self, module_name):
        if module_name[0] == '$':
            if module_name[1:] in self.runnable_controller.all_runnable_module.keys():
                return module_name[1:]
            return self.create_runnable_module(module_name[1:], self.runnable_controller._runnable_define[module_name[1:]])
        elif module_name[0] == '*':
            return self.create_runnable_module(module_name[1:], RunnableModule.get_default_config(module_name[1:]))
        else:
            return self.create_runnable_module(module_name, self.runnable_controller._runnable_define[module_name])

    def module_parser(self, name):
        name = name.strip()
        if '(' in name:
            modules = [i for i in re.split(r'\((.*)\)', name) if len(i) > 0]
            if len(modules)==1:
                return self.add_from_script(modules[0])
            elif len(modules)==2:
                sub_names = self.add_from_script(modules[1])
                name = self.create_module(modules[0])
                self.runnable_controller.all_runnable_module[name].required.update(set(sub_names))
                return [name]
            else:
                raise NotImplementedError("{}".format(modules))
        else:
            return [self.create_module(name)]

    def add_from_script(self, script):
        runnable_modules = [i.strip() for i in self.parsing(script)]
        module_names = []
        pre = None
        for module in runnable_modules:
            outs = self.module_parser(module)
            module_names += outs
            if pre is not None:
                self.connect_edge(pre, outs[0])
            pre = outs[-1]
        return module_names

    def create_runnable_module(self, name, config):
        if name in RunnableModule.get_cls_count().keys():
            RunnableModule.get_cls_count()[name] += 1
            name = "{}_{}".format(name, RunnableModule.get_cls_count()[name])
        else:
            RunnableModule.get_cls_count()[name] = 0

        current_module = sys.modules[__name__]
        config = self.get_argument(config)

        self.runnable_controller.all_runnable_module[name] = getattr(current_module, config.command)(name, config)
        return name

    def get_argument(self, command_arg):
        """
        [Configuration safe]
        Create a new argment that inherits the parent property of argmemt.
        :param command_arg:
        :return:
        """
        arg = Config.get_empty()
        if 'inherit' in command_arg.keys():
            arg = self.get_argument(self.runnable_controller._runnable_define[command_arg.inherit])
        arg.update(command_arg)
        return arg

class RunnableModuleController(object):
    def __init__(self, config, controller):
        self.config = config

        self.next = defaultdict(list)
        self.previous = defaultdict(list)

        self.make_directory()
        self.factory = RunnableModuleFactory(self, config.ipc_host, int(config.ipc_port))
        RunnableModule.set_global_module(controller, self)

        self.command_configs = Config.from_yaml(self.config.command_path)
        self._runnable_define = self.command_configs.DEFINE

        # Shared memory: self.variables
        self.variables = defaultdict(None)

        self.variables['main'] = ['root']
        self.variables['iter'] = ['root']
        self.variables['self'] = ['root']

        root_config = RunnableModule.get_default_config('RunnableGraph')
        root_config.repeat = 1
        self.root: RunnableModule = RunnableGraph(config=root_config, name='root')
        self.all_runnable_module = {'root': self.root}

        for script in self.command_configs.DEFAULT:
            self.factory.job_queue.put(('ADD', script))

        self.factory.start()
        #DEBUG
        # for key in self.all_runnable_module.keys():
        #     print(key, self.all_runnable_module[key].indegree)
        #     print(self.all_runnable_module[key].required)
        #     print('adj', self.next[key])
        # input()

    def get_runnable_module(self, key):
        """
        :param key:
        This function supports variable type arguments.
        In the case of $, the value is taken from self.vairables and used as a key.
        If self.variables is a list, select the last value.
        :return:
        """
        try:
            if key[0]=='$':
                key = self.get_last_variables_value(key[1:])
            return self.all_runnable_module[key]
        except AttributeError:
            return None
        except KeyError:
            raise RunnableModuleDestroyException()

    def remove_runnable_module(self, key):
        del self.all_runnable_module[key]

    def make_directory(self):
        workspace = self.config.get('save_path', '$base/ipc', possible_none=False)
        App.instance().make_save_dir(workspace)

    def get_last_variables_value(self, name):
        item = self.variables[name]
        if isinstance(item, list):
            item = item[-1]
        return item

    def get_current_main_module(self):
        """
        return main module.
        The object of this function must be used only as a local variable.
        """
        return self.get_runnable_module(self.variables['main'][-1])

    def run(self):
        def loading():
            while True:
                for i in range(4): yield '.' * i

        temp_bar = loading()
        while True:
            try:
                self.root.step = 0
                self.root.run()
                while len(self.root.required)==0 and self.factory.job_queue.empty():
                    print('\ridle' + next(temp_bar), end='')
                    time.sleep(0.5)
            except Exception as e:
                if App.instance().config.App.DEBUG:
                    print(traceback.format_exc())
                self.factory.end()
                raise e




from framework.app.app import App, SingletoneInstance
from pyparsing import nestedExpr
import re
import os

class Formatter():
    def __init__(self, runnable_controller, contents={}, format=''):
        """
        Manages various formats.
        This class is dependent on the current Controller and configuration safe.

        :param controller: Main Controller
        :param contents: Variables dictionary
        :param format: The format variable must be specified in [$name] format.
        example:
            f = Formatter(Controller, contents={'content':'model', 'format':'exe'}, format='[$content]_checkpoint.[$format]')
            print(f.Formatting())
            -> model_checkpoint.exe
        """
        self.runnable_controller = runnable_controller
        self.contents = contents
        self.format = format

    def StateFormatting(self, variable):
        """
        Create a string according to the given format. Variables are separated by $.
        If the variable contains :, it means that the variable of state is printed.
        it consist of [$STATE_NAME:VARIABLE_NAME:ZFILL]
        If it is $main:step:05, the step variable of the main state is output as zfill(5).

        STATE_NAME: main or now
        VARIABLE_NAME: state variable name
        ZFILL : How many zeros to fill

        :param variable:
        :return:
        """
        try:
            format = variable.split(':')
            if len(format)==1:
                return self.contents[variable[1:]] if variable[0]=='$' else variable
            state = format[0]
            var = format[1]
            zfill = 1
            if len(format)>2:
                zfill = int(format[2])
            module = self.runnable_controller.get_runnable_module(state)
            return str(module.__getattribute__(var)).zfill(zfill)
        except Exception as e:
            raise e

    def Formatting(self) -> str:
        """
        Assign variables according to format. return string
        :return: formatted str
        """
        try:
            out = ''
            formats = [self.StateFormatting(i) for i in re.split(r"\[(.*?)\]", self.format) if len(i)>0]
            for i in formats:
                out += str(i)
            return out
        except Exception as e:
            raise e

    def MakeContentsFromString(self, string: str) -> None:
        """
        Extracts a variable according to the format from the given string.
        The outside of the format must be independent.
        :param string:
        :return:
        """
        try:
            formats = [i for i in re.split(r"\[(.*?)\]", self.format) if len(i)>0]
            comparator = [f for f in formats if f[0] != '$']
            formatter = [f for f in formats if f[0] == '$']
            count = 0
            for comp in comparator:
                self.contents[formatter[count][1:]] = string[:string.find(comp)]
                string = string[string.find(comp)+len(comp):]
                count += 1
            self.contents[formatter[count][1:]] = string

        except Exception as e:
            raise e

    def CheckSameBaseName(self, f1, f2) -> bool:
        """
        Check if two elements are the same except for format
        :param f1: string 1
        :param f2: string 2
        :return: true if all elements except format are the same

        """
        try:
            formats = [i for i in re.split(r"\[(.*?)\]", self.format) if len(i)>0]
            comparator = [f for f in formats if f[0] != '$']
            return all([i in f1 and i in f2 for i in comparator])
        except Exception as e:
            raise e

class MainStateBasedFormatter(Formatter):
    def __init__(self, runnable_controller, contents={}, format=''):
        super().__init__(runnable_controller, contents, format)

if __name__ == '__main__':
    f = Formatter(None, contents={'content':'cdef','batch':523,'format':'exe'},format='[$content]_[$batch]_checkpoint.[$format]')
    a = f.Formatting()
    f = Formatter(None, contents={'content':'asdasd','batch':523,'format':'exe'},format='[$content]_[$batch]_checkpoint.[$format]')
    b = f.Formatting()
    print(f.CheckSameBaseName(a,b))


'''
    @staticmethod
    def variable_name_parsing(line, separator='$'):
        i = 0
        if line[0] == separator:
            i+=1
        temp = ''
        while i < len(line):
            if not (temp + line[i]).isidentifier():
                break
            temp += line[i]
            i+=1
        return temp, i
'''


import numpy
from framework.dataloader.TensorTypes import *

"""
Basic transform methods
"""

class Transforms(object):
    def __init__(self, required=[], meta={}):
        self.required = required
        self.meta = meta
        if 'ALL' in self.meta.keys():
            for name in self.required:
                self.meta[name] = self.meta['ALL']

class ToTensor(Transforms):
    def __init__(self, required=[], meta={}):
        super(ToTensor, self).__init__(required, meta)

    def __call__(self, samples):
        for sample in self.required:
            data = samples[sample]
            meta_data = self.meta[sample]

            if isinstance(meta_data, IMAGE):
                samples[sample] = data.transpose((2, 0, 1))
            else:
                raise NotImplementedError
        return samples

class ToNumpy(Transforms):
    def __init__(self, required=[], meta={}):
        super(ToNumpy, self).__init__(required, meta)

    def __call__(self, samples):
        for sample in self.required:
            data = samples[sample]
            meta_data = self.meta[sample]
            if isinstance(meta_data, IMAGE):
                samples[sample] = data.permute((0,2,3,1)).cpu().detach().numpy()
            else:
                raise NotImplementedError
        return samples



from framework.utils.config import Config
import numpy as np
import logging
import logging.config
import datetime
import sys
import copy
import torch
import os

DEFAULT_CONFIG_PATH = "{}/default_config.yaml".format(os.path.dirname(os.path.abspath( __file__ )))
class SingletoneInstance:
    __instance = None

    @classmethod
    def __getInstance(cls):
        return cls.__instance

    @classmethod
    def instance(cls, *args, **kargs):
        cls.__instance = cls(*args, **kargs)
        cls.instance = cls.__getInstance # Override
        return cls.__instance

    @classmethod
    def reinitialize(cls, *args, **kargs):
        cls.__instance = cls(*args, **kargs)

    @classmethod
    def assign_instance(cls, instance):
        cls.__instance = instance

''':key
$
'''
class App(SingletoneInstance):
    @classmethod
    def make_from_config_list(cls, config_paths):
        """
        Initialize the App by using the configuration list. This function must be executed only once.
        :param config_paths:
        :return: App Class
        """
        config = Config.get_empty()

        for path in config_paths:
            config.update(config.from_yaml(path))
        return cls.instance(config, update=True)

    def __init__(self, config=None, update=False):
        """
        App class [Configuration unsafe]

        App is responsible for the overall structure of this framework.
        App has all configurations. please check the list.
        - App
        - MODEL_CONTROLLER:
            - COMMAND_CONTROLLER
            - MODEL_STATE_CONTROLLER
            - OPTIMIZER
            - LOSSES
        - DATA_LOADER
        - SYSTEM

        App can manage environment variables through variables.
        When using variables, use $variable_name.

        :param config:
        :param update: If update is true, all System parameters are executed. (Logger, Seed...)
        """
        self.config = Config.from_yaml(DEFAULT_CONFIG_PATH)
        if config is not None:
            self.config.update(config)

        # --------------------------------SYSTEM--------------------------------------
        self.system = self.config.SYSTEM
        logging.config.dictConfig(Config.extraction_dictionary(self.system.LOGGER))
        self.logger = logging.getLogger('DEFAULT')
        self.setSeed(self.system.SEED)


        # --------------------------------Variable Setting--------------------------------------
        self.name = self.config.App.NAME
        self.mode = self.config.App.mode
        self._variables = self.config.App.Variables

        base = self._variables.directory_root
        model_name = self.name
        experiment_name = self.config.App.EXPERIMENT_NAME

        if self.mode == 'finetuning':
            if os.path.exists(os.path.join(base, model_name, experiment_name)):
                self._variables.pre = os.path.join(base, model_name, experiment_name)
                experiment_name = self.name_format(experiment_name)
            base_path = os.path.join(base, model_name, experiment_name)
            self._variables.base = base_path
            if os.path.exists(os.path.join(base, model_name, experiment_name)):
                print('[INFO] Now App mode is {}, Since the directory already exists, a new folder {} is created.'.format(self.mode,
                                                                                                                          self._variables.base))
        elif self.mode == 'new':
            if os.path.exists(os.path.join(base, model_name, experiment_name)):
                experiment_name = self.name_format(experiment_name)
            base_path = os.path.join(base, model_name, experiment_name)
            self._variables.base = base_path
            if os.path.exists(os.path.join(base, model_name, experiment_name)):
                print('[INFO] Now App mode is {}, Since the directory already exists, a new folder {} is created.'.format(
                    self.mode, self._variables.base))

        elif self.mode == 'resume':
            if not os.path.exists(self._variables.base):
                raise FileExistsError
        else:
            raise NotImplementedError
        print('[INFO] base_path: {}'.format(self._variables.base))

    def setSeed(self, seed):
        """
        set seed..
        :param seed:
        """
        torch.manual_seed(seed)
        np.random.seed(seed)

        if self.system.REPRODUCIBILITY:
            self.logger.info('REPRODUCIBILITY MODE TRUE')
            torch.backends.cudnn.deterministic = True
            torch.backends.cudnn.benchmark = False

    def name_format(self, name: str)-> str:
        """
        :param name:
        :return: Returns the combined names of common time formats
        """
        now = datetime.datetime.now()
        return now.strftime('{}_%y%m%d_%Hh%Mm'.format(name))

    def time_format(self) -> object:
        """
        :return: Returns common time formats
        """
        now = datetime.datetime.now()
        return now.strftime('%y%m%d-%H:%M:%S')

    def get_device(self):
        """
        :return: Returns device name
        """
        device_name = self.config.App.device
        return device_name

    def get_gpu_ids(self)->list:
        """
        Return gpu ids
        :return: gpu idxes
        """
        gpu_ids = self.config.App.gpu_ids
        return gpu_ids

    def set_gpu_device(self, module: torch.nn.Module) -> torch.nn.Module:
        """
        Set the device in the module. Currently, only the Data parallel method is supported.

        :param module:
        :return:
        """
        device_name = self.config.App.device
        gpu_ids = self.config.App.gpu_ids
        parallel = self.config.App.parallel

        device = torch.device('cpu' if device_name is 'cpu' else 'cuda')

        if device_name is not 'cpu' and len(gpu_ids) > 1 and parallel is "data":
            return torch.nn.DataParallel(module, gpu_ids).to(device)
        return module.to(device)

    def get_base_path(self):
        """
        Return the workspace.
        :return:
        """
        return self.get_variables('$base')

    def make_save_dir(self, path: str):
        """
        After parsing a variable, convert it to the entire path and create a directory.
        This function affects os (create file)
        :param path:
        """
        os.makedirs(self.dir_path_parser(path), exist_ok=True)

    @staticmethod
    def dir_path_parser(path):
        return ('/' if path.startswith('/') else '') + os.path.join(*App.instance().variable_parsing(path, '/'))

    def smart_write(self, contents, dst, mode=None):
        """
        tepmly
        :param contents:
        :param dst:
        :param mode:
        """
        if isinstance(dst, str):
            if dst == 'stdout':
                sys.stdout.write(contents)
            else:
                with open(dst, mode) as f:
                    f.write(contents)
        else:
            with open(dst, mode) as f:
                f.write(contents)

    def set_variables(self, variable, value):
        if variable[0] != '$':
            raise Exception('Variable name must to start with $')
        self._variables[variable[1:]] = value

    def get_variables(self, variable):
        if variable[0] != '$':
            raise Exception('Variable name must to start with $')
        return self._variables[variable[1:]]

    def variable_parsing(self, string, sep=' '):
        """
        Assigns all variables in the given string. variable is defined as $.
        This function returns the cloned result. [safe from shallow copy]
        :param string:
        :param sep:
        :return:
        """
        return [copy.deepcopy(self.get_variables(p)) if p[0] == '$' else p for p in string.split(sep) if len(p)>0]



import torch
from framework.utils.runtime import get_class_object_from_name
from framework.utils.config import Config
from framework.app.Format import *
import os

class BaseModel(torch.nn.Module):
    def __init__(self, config):
        super().__init__()
        self.config = config
        self.inputs_name = []
        self.outputs_name = []

    def description(self):
        return str(self) + "=> IN:{}, OUT:{} ".format(
            self.inputs_name, self.outputs_name
        ) + "device: {}:{}\n".format(App.instance().get_device(), App.instance().get_gpu_ids())

    @classmethod
    def model_factory(cls, config_path):
        config = Config.from_yaml(config_path)

        dataset_module = config.module_name
        dataset_class = config.class_name
        dataset_module: BaseModel = get_class_object_from_name(dataset_module, dataset_class)
        return dataset_module(config)

    def forward(self, samples):
        for name in self.inputs_name:
            samples[name] = samples[name].to(device=App.instance().get_device()).float()
        return samples

    def get_name_format(self, controller):
        f = MainStateBasedFormatter(controller, {'model_name': App.instance().name, 'time': App.instance().time_format()},'[$time]_[$model_name]_[$main:step:03]_[$main:total_step:08].model')
        return f.Formatting()

    def save(self, info):
        dst = info['path']
        App.instance().make_save_dir(dst)

        file_names = os.listdir(dst)
        previous_file_name = None
        for name in file_names:
            _, file_extension = os.path.splitext(name)
            if file_extension == 'model':
                epoch = name.split('_')[-2]
                if int(epoch) == info['controller'].get_current_main_module():
                    previous_file_name = name

        saved_data = {'data': self.state_dict(), 'config': Config.extraction_dictionary(self.config)}
        saved_path = os.path.join(dst, self.get_name_format(info['controller']))
        torch.save(saved_data, saved_path)
        if previous_file_name is not None:
            os.remove(os.path.join(dst, previous_file_name))

        App.instance().set_variables('$latest_{}'.format(info['module_name']), saved_path)

    def load(self, info):
        path = info['path']
        state_dict = torch.load(path)
        if info.get('config_load', False):
            del self.config
            self.config = Config.from_dict(state_dict['config'])
        self.load_state_dict(state_dict['data'], strict=info['load_strict'])



import matplotlib.pyplot as plt
import copy
from framework.app.Format import *
from collections import defaultdict
from framework.ipc.ThreadCommand import *
from framework.utils.config import Config
from framework.dataloader.DataLoader import DataLoaderController
import traceback
from framework.app.Exceptions import *
from tqdm import tqdm
import os
from framework.dataloader.Transform import TRANSFORM
from framework.dataloader.TensorTypes import *


class DependentParent(object):
    def __init__(self, my, run_cycle: str):
        abb = run_cycle.split(':')
        default_step = 1
        if len(abb) > 1:
            run_cycle = abb[0]
            default_step = int(abb[1])

        cut = run_cycle.split('.', maxsplit=1)
        if len(cut)==1: cut.append("step")

        self.my = my
        self.dependent_module, self.attribute = cut
        self.requried_step = default_step
        self.previous_step = -1

    def get_dependent_module(self):
        return self.my.get_runnable_controller().get_runnable_module(self.dependent_module)

    def get_condition_value(self):
        return self.my.get_runnable_controller().get_runnable_module(self.dependent_module).__getattribute__(self.attribute)

    def valid_check(self):
        now_step = self.get_condition_value()
        if now_step % self.requried_step == 0 and self.previous_step != now_step and self.my.live:
            self.previous_step = now_step
            return self.my.repeat == -1 or self.my.repeat > self.my.step
        return False

class RunnableModule(object):
    __cls_counter = defaultdict(int)
    __model_controller = None
    __runnable_controller = None
    __factory = None
    __global_count = 0

    @classmethod
    def get_default_config(cls, command_name):
        args = {}
        args['command'] = command_name
        args['required'] = []
        args['repeat'] = 1
        args['args'] = {}
        return Config.from_dict(args)

    @property
    def factory(self):
        return RunnableModule.__factory

    @classmethod
    def set_global_module(cls, model_controller, runnable_controller):
        cls.__model_controller = model_controller
        cls.__runnable_controller = runnable_controller
        cls.__factory = runnable_controller.factory

    @classmethod
    def get_cls_count(cls):
        return cls.__cls_counter

    @classmethod
    def get_model_controller(cls):
        return cls.__model_controller

    @classmethod
    def get_runnable_controller(cls):
        return cls.__runnable_controller

    def write_log(self, contents, **kwargs):
        latest_state_obj = self.get_runnable_controller().get_current_main_module()
        if 'tqdm' not in dir(latest_state_obj) or latest_state_obj is None:
            print(contents, **kwargs)
            return
        if latest_state_obj.tqdm is None:
            print(contents, **kwargs)
        else:
            latest_state_obj.tqdm.write(contents)

    def __init__(self, name, config):
        """

        :param name:
        :param config: must to be configuration safe
        :param generation:
        :param parent:
        """
        self.config = config
        self.required = set(self.config.get('required', []))

        self.name = name
        self.step = 0
        self.finish = self.config.get('finish', False)
        self.live = True

        # Graph Module
        self.indegree = 0
        self.current_degree = 0

        self.write_log("[COMMAND JOIN] ModuleName:{}, Command:{}".format(self.name, self.__class__.__name__))
        """
        Command Class
        Command class can be used with the with statement.
        :param name:
        :param config:
        """
        self.dependent = DependentParent(self, self.config.get('run_cycle', "$self.step:1"))
        self.repeat = int(self.config.get('repeat', 1, False))
        if not 'args' in self.config.keys():
            self.config.args = {}

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.current_degree = 0

        if exc_type is not None:
            self.config.ERROR = Exception(exc_type, exc_val, exc_tb)
#            raise Exception(exc_type, exc_val, exc_tb)

    def __enter__(self):
        return self

    def start(self):
        """
        Must be executed before the module is executed.
        """
        RunnableModule.get_runnable_controller().variables['self'].append(self.name)

    def valid(self):
        return self.dependent.valid_check()

    def run(self):
        """
        Must be executed before the module is executed.
        """
        raise NotImplementedError

    def __del__(self):
        self.write_log("[COMMAND EXPIRED] ModuleName:{}, Command:{}".format(self.name, self.__class__.__name__))

    def update(self):
        """
        Must be executed before the module is executed.
        """
        raise NotImplementedError

    def end(self):
        self.current_degree = 0

        if RunnableModule.get_runnable_controller().variables['self'][-1] == self.name:
            RunnableModule.get_runnable_controller().variables['self'].pop()

        if self.finish and (self.repeat != -1 and self.repeat <= self.step):
            self.live = False

    def leave(self):
        return self.live

    def destroy(self):
        raise NotImplementedError

class RunnableGraph(RunnableModule):
    def __init__(self, name, config, **kwargs):
        super().__init__(name, config)
        self.process_queue = queue.Queue()

    def start(self):
        super(RunnableGraph, self).start()

    def end(self):
        super(RunnableGraph, self).end()

    def init(self):
        self.process_queue = queue.Queue()
        self.step += 1

        # UPDATE GRAPH
        self.factory.update_graph()
        for obj_name in self.required:
            obj = RunnableModule.get_runnable_controller().get_runnable_module(obj_name)
            obj.step = 0
            obj.current_degree = 0
            if obj.indegree == 0:
                self.process_queue.put(obj_name)

    def update(self):
        pass

    def destroy(self):
        module_controller = RunnableModule.get_runnable_controller()
        for obj_name in self.required:
            module_controller.get_runnable_module(obj_name).destroy()
        module_controller.remove_runnable_module(self.name)

    def remove_edge(self, name):
        module_controller = RunnableModule.get_runnable_controller()
        next_names = module_controller.next[name]
        previous_names = module_controller.previous[name]

        for next_name in next_names:
            next_obj = module_controller.get_runnable_module(next_name)
            next_obj.indegree -= 1
            next_obj.indegree += len(previous_names)
            next_obj.current_degree += len(previous_names)

            module_controller.previous[next_name].remove(name)
            module_controller.previous[next_name] += previous_names
            if next_obj.current_degree == next_obj.indegree:
                self.process_queue.put(next_name)

        for pname in previous_names:
            module_controller.next[pname].remove(name)
            module_controller.next[pname] += next_names

        del module_controller.next[name]
        del module_controller.previous[name]
        self.required.remove(name)

    def process(self, obj_name):
        module_controller = RunnableModule.get_runnable_controller()
        try:
            with module_controller.get_runnable_module(obj_name) as p:
                if p.valid():
                    p.start()
                    p.run()
                    p.update()
        except MainGraphStepInterrupt as e:
            if isinstance(self, MainGraph):
                pass
            else:
                p.end()
                if not p.leave():
                    self.remove_edge(p.name)
                    p.destroy()
                raise e
        except MainGraphFinishedException as e:
            p.end()
            if not p.leave():
                self.remove_edge(p.name)
                p.destroy()
            raise e
        except Exception as e:
            if App.instance().config.App.DEBUG:
                print(traceback.format_exc())
            else:
                self.write_log('[ERROR/{}/{}] {}'.format(p.name, p.__class__.__name__, e))
            pass
        finally:
            p.end()
            if not p.leave():
                self.remove_edge(p.name)
                p.destroy()
            else:
                nexts_name = module_controller.next[p.name]
                for name in nexts_name:
                    next_obj = module_controller.get_runnable_module(name)
                    next_obj.current_degree += 1
                    if next_obj.current_degree == next_obj.indegree:
                        self.process_queue.put(name)

    # Must to SCC
    def run(self):
#        self.write_log("[START] -> {} {}/{}, task={}".format(self.name, self.step, self.config.repeat, self.required))
        while self.valid():
            self.init()
#            self.write_log("[PROCESSING] -> {} {}/{}, task={}".format(self.name, self.step, self.config.repeat, self.required))
            while not self.process_queue.empty():
                obj_name: str = self.process_queue.get()
                self.process(obj_name)
#        self.write_log("[END] -> {} {}/{}".format(self.name, self.step, self.config.repeat))

class MainGraph(RunnableGraph):
    def __init__(self, name, config, **kwargs):
        super().__init__(name, config, **kwargs)
        self.iter = None
        self.tqdm = None

        self.total_step = 0
        self.loader_name = self.config.args.loader_name

    def get_epoch(self):
        return self.step - 1

    def start(self):
        super(MainGraph, self).start()
        RunnableModule.get_runnable_controller().variables['main'].append(self.name)

        config = copy.deepcopy(self.config.args[self.loader_name]) if self.loader_name in self.config.args.keys() else None
        DataLoaderController.instance().make_dataset(self.loader_name, config)
    
    def init(self):
        super(MainGraph, self).init()
        self.tqdm = tqdm(DataLoaderController.instance().dataloaders[self.loader_name],
                                       bar_format='{l_bar}{bar:10}{r_bar}', ascii=True)
        self.tqdm.desc = '[{}] {}/{} '.format(self.loader_name, self.step, self.repeat)
        self.iter = iter(self.tqdm)

    def destroy(self):
        super(MainGraph, self).destroy()
        self.tqdm = None
        self.iter = None

    def end(self):
        super(MainGraph, self).end()
        if RunnableModule.get_runnable_controller().variables['main'][-1] == self.name:
            RunnableModule.get_runnable_controller().variables['main'].pop()

    def run(self):
        try:
            super(MainGraph, self).run()
        except MainGraphFinishedException as e:
            self.write_log('[INTERRUPT/{}/{}] {}'.format(self.name, self.__class__.__name__, e))

# -------------------------------- NODE
class RunnableNode(RunnableModule):
    def __init__(self, name, config):
        super().__init__(name, config)

    def start(self):
        pass

    def run(self):
        raise NotImplementedError

    def update(self):
        self.step += 1

    def end(self):
        super(RunnableNode, self).end()

    def destroy(self):
        module_controller = RunnableModule.get_runnable_controller()
        module_controller.remove_runnable_module(self.name)

class TestCommand(RunnableNode):
    def __init__(self, name, config):
        super().__init__(name, config)

    def run(self):
        new_step = int(self.config.args.new_step)
        obj = self.get_runnable_controller().get_current_main_module()
        previous_step = obj.step
        obj.step = new_step
        self.write_log('previous step: {} -> new step {}'.format(previous_step, new_step))
        if not obj.valid():
            raise MainGraphFinishedException

class PrintCommand(RunnableNode):
    def __init__(self, name, config):
        config = Config.from_dict({
            'args':{
                'content': 'default contents'
            }
        }).update(config)
        super().__init__(name, config)

    def run(self):
        self.write_log('[{}] PRINT MODULE: {} step: {} / repeat: {}'.format(self.name, self.config.args.content,
                                                                            self.dependent.get_dependent_module().__getattribute__(self.dependent.attribute), self.repeat), end=' ')
        for name in self.required:
            self.write_log(name, end=" ")
#            self.write_log(name + " " + controller.sample[name])

class PrintStateCommand(RunnableNode):
    def __init__(self, name, config):
        config = Config.from_dict({
            'args':{
                'obj': '$main'
            }
        }).update(config)
        super().__init__(name, config)

    def run(self):
        obj = self.get_runnable_controller().get_runnable_module(self.config.args.obj)
        for name in self.required:
            self.write_log('[{}.{}]: {} / '.format(obj.name, name, obj.__getattribute__(name)), end=' ')
        self.write_log('')

class RunCommand(RunnableNode):
    def __init__(self, name, config):
        super().__init__(name, config)

    def run(self):
        for name in self.required:
            if name in self.get_model_controller().all_callable:
                self.get_model_controller().__getattribute__(name)(**self.config.args)

class BatchedImageShowCommand(RunnableNode):
    def __init__(self, name, config):
        super().__init__(name, config)
        self.numpy_trasnsform = TRANSFORM['ToNumpy']([name + "_output" for name in self.config.required], {'ALL': IMAGE()})

    def run(self):
        temp_sample = {}
        for name in self.config.required:
            temp_sample[name + "_output"] = self.get_model_controller().sample[name]
        temp_sample = self.numpy_trasnsform(temp_sample)

        for key in temp_sample.keys():
            plt.imshow(temp_sample[key][self.config.args.batch_number])
            plt.show()

class BatchedImageSaveCommand(RunnableNode):
    def __init__(self, name, config):
        super().__init__(name, config)
        self.numpy_trasnsform = TRANSFORM['ToNumpy']([name for name in self.config.required], {'ALL': IMAGE()})
        self.type = IMAGE()

    def leave(self):
        if self.live or not MultipleProcessorController.instance().finished(self.__class__.__name__):
            return True
        else:
            return False

    def destroy(self):
        MultipleProcessorController.instance().remove_process(self.__class__.__name__)

    def run(self):
        args = []
        dir_path = App.instance().dir_path_parser(self.config.args.get('path', '$base/visual/img', possible_none=False))
        fm = self.config.args.get('format', 'png', possible_none=False)

        App.instance().make_save_dir(dir_path)
        formatter = MainStateBasedFormatter(self.get_runnable_controller(), {'content': '', 'format': fm, 'batch':0},
                              format='[$main:step:03]e_[$main:total_step:08]s_[$content]_[$batch].[$format]')

        for name in self.config.required:
            formatter.contents['content'] = name
            imgs = self.numpy_trasnsform({name:self.get_model_controller().sample[name].clone()})

            b,_,_,_ = imgs[name].shape
            for i in range(b):
                formatter.contents['batch'] = str(i).zfill(4)
                path = os.path.join(dir_path, formatter.Formatting())
                args.append((path, imgs[name][i]))

        import time
        def batched_image_save(queue):
            while True:
                sample = queue.get()
                if sample is None: break
                path, img = sample
                misc.imsave(path, img)
                time.sleep(0.001)
        MultipleProcessorController.instance().push_data(self.__class__.__name__, batched_image_save, args, num_worker=1)


class ModuleLoadClass(RunnableNode):
    def __init__(self, name, config):
        super().__init__(name, config)

    def run(self):
        for name in self.required:
            args = Config.from_dict({
                'module_name': name,
                'controller': self.get_runnable_controller(),
                'path': '$latest',
                'load_strict': True
            })
            args.update(self.config.args[name])
            try:
                if args['path'] == '$latest':
                    args['path'] = args['path'] + '_{}'.format(name)
                args['path'] = App.instance().dir_path_parser(args['path'])
            except KeyError as e:
                self.write_log('[ERROR] App.variables load error. key:{}.'.format(args['path']))
                continue

            if name not in dir(self.get_model_controller()):
                self.write_log('[ERROR] Load fail: {} is not Modules.'.format(name))
                continue

            module = self.get_model_controller().__getattribute__(name)
            if not isinstance(module, torch.nn.Module):
                self.write_log('[ERROR] Load fail: {} is not Modules'.format(name))
                continue
            if not callable(getattr(module, 'load')):
                self.write_log('[ERROR] Load fail: {} must to have load method.'.format(name))
                continue
            module.load(args)
            self.write_log('[INFO] {} Load scucesses [{}] '.format(self.config.required, args['path']))

class ModuleSaveClass(RunnableNode):
    def __init__(self, name, config):
        super().__init__(name, config)

    def run(self):
        for name in self.required:
            if name not in dir(self.get_model_controller()):
                self.write_log('[ERROR] save fail: {} is not Modules.'.format(name))
                continue
            args = Config.from_dict({
                'module_name': name,
                'controller': self.get_runnable_controller(),
                'path': '$base/ckpt_{}'.format(name),
                'load_strict': True
            })
            args.update(self.config.args[name])
            args['path'] = App.instance().dir_path_parser(args['path'])

            module = self.get_model_controller().__getattribute__(name)
            if not isinstance(module, torch.nn.Module):
                self.write_log('[ERROR] save fail: {} is not Modules'.format(name))
                continue
            if not callable(getattr(module, 'save')):
                self.write_log('[ERROR] save fail: {} must to have save method.'.format(name))
                continue

            module.save(args)
            self.write_log('[INFO] {} Save scucesses [{}] '.format(self.config.required, App.instance().get_variables('$latest_{}'.format(name))))





import torch
import scipy.misc as misc

class TensorType:
    def __str__(self):
        return self.__class__.__name__

    def getSample(self, sample):
        raise NotImplementedError

class IMAGE(TensorType):
    def image_loader(self, path):
        return misc.imread(path)/255.

    def image_saver(self, path, data):
        return misc.imsave(path, data)

    def getSample(self, sample):
        return self.image_loader(sample)

class GRAYSCALE_IMAGE(IMAGE):
    def image_loader(self, path):
        return misc.imread(path)

    def getSample(self, sample):
        return self.image_loader(sample)


class DISPARITY(TensorType):
    pass




import threading
import queue
from framework.ipc.Command import Command
from framework.app.app import SingletoneInstance
from multiprocessing import Queue, Process
import time

class ProcessWrapper(object):
    def __init__(self, func, worker_num=8):
        self.queue = Queue()
        self.worker_num = worker_num
        self.processes = [Process(target=func, args=(self.queue,)) for _ in range(self.worker_num)]

    def run(self):
        for p in self.processes: p.start()

    def stop(self):
        for _ in range(self.worker_num): self.queue.put(None)
        while not self.queue.empty(): time.sleep(0.5)
        for p in self.processes: p.join()

    def __del__(self):
        print('ProcessWarpper del')

class MultipleProcessorController(SingletoneInstance):
    def __init__(self):
        self.processes = {}

    def remove_all_process(self):
        for name in self.processes.keys():
            p = self.processes.get(name, None)
            if p is None: continue
            self.remove_process(name)

    def finished(self, name):
        p = self.processes.get(name, None)
        if p is None:
            return True
        return self.processes[name].queue.empty()

    def remove_process(self, name):
        p = self.processes.get(name, None)
        if p is None:
            return
        p.stop()
        self.processes[name] = None

    def make_new_process(self, name, func, num_worker=8):
        self.processes[name] = ProcessWrapper(func, num_worker)

    def push_data(self, name, func, datas, num_worker=8):
        if self.processes.get(name, None) is None:
            self.make_new_process(name, func, num_worker=num_worker)
            self.processes[name].run()

        process: ProcessWrapper = self.processes[name]
        for data in datas:
            process.queue.put(data)




import torch
from framework.app.app import App
from framework.utils.runtime import get_class_object_from_name
from framework.utils.config import Config
import os

def make_optimizer(args, model):
    '''
    ref: https://github.com/thstkdgus35/EDSR-PyTorch/blob/master/src/utility.py
    '''

    trainable = filter(lambda x: x.requires_grad, model.parameters())
    optimizer_class = get_class_object_from_name(args.optimizer.optimizer_module,
                                                 args.optimizer.optimizer_class)
    optimizer_args = Config.extraction_dictionary(args.optimizer.optimizer_args)

    scheduler_class = get_class_object_from_name(args.scheduler.scheduler_module,
                                       args.scheduler.scheduler_class,)
    scheduler_args = Config.extraction_dictionary(args.scheduler.scheduler_args)

    class CustomOptimizer(optimizer_class):
        def __init__(self, *args, **kwargs):
            super(CustomOptimizer, self).__init__(*args, **kwargs)

        def _register_scheduler(self, scheduler_class, **kwargs):
            self.scheduler = scheduler_class(self, **kwargs)

        def save(self, info):
            dst = info['path']
            file_names = os.listdir(dst)
            previous_file_name = None
            for name in file_names:
                _, file_extension = os.path.splitext(name)
                if file_extension == 'opt':
                    epoch = name.split('_')[-2]
                    if int(epoch) == info['state'].epoch:
                        previous_file_name = name

            torch.save(self.state_dict(), os.path.join(dst, self.get_save_name(info['state'])))
            if previous_file_name is not None:
                os.remove(os.path.join(dst, previous_file_name))

        def load(self, load_dir, epoch=1):
            self.load_state_dict(torch.load(self.get_dir(load_dir)))
            if epoch > 1:
                for _ in range(epoch): self.scheduler.step()

        def schedule(self):
            self.scheduler.step()

        def get_lr(self):
            return self.scheduler.get_lr()[0]

        def get_last_epoch(self):
            return self.scheduler.last_epoch

        def get_save_name(self, state):
            return App.instance().name_format(App.instance().name) + "_{}_{}.opt".format(state.epoch, state.step)

    optimizer = CustomOptimizer(trainable, **optimizer_args)
    optimizer._register_scheduler(scheduler_class, **scheduler_args)
    return optimizer



from framework.app.app import SingletoneInstance, App
from framework.dataloader.DataLoader import DataLoaderController
from framework.utils.runtime import get_class_object_from_name
from framework.trainer.losses import LossContainer
from framework.ipc.RunningGraphController import RunnableModuleController
from tqdm import tqdm
import torch
import time
''': ModelState

ModelState: START -> EPOCH_START -> BATCH_START -> BATCH_END -> EPOCH_END -> TERMINATE
                          |-----------------------------------------|
'''

class ModelState(object):
    def __init__(self, name='START', step=1, epoch=0, loader_name='global'):
        self.state_name = name
        self.step = step
        self.epoch = epoch
        self.loader_name = loader_name
        self.iter = None
        self.tqdm: tqdm = None

class ModelController(SingletoneInstance):
    def __init__(self):
        self.config = App.instance().config.MODEL_CONTROLLER
        self.dataloader_controller = DataLoaderController.instance()

        self.MODEL: torch.nn.Module = None
        self.OPTIMIZER: torch.nn.Module = None
        self.LOSSES: torch.nn.Module = App.instance().set_gpu_device(LossContainer(self.config.LOSSES))

#        self.COMMAND_CONTROLLER = CommandController(self.config.COMMAND_CONTROLLER.command_path, self)
        self.COMMAND_CONTROLLER = RunnableModuleController(self.config.COMMAND_CONTROLLER, self)
        self.sample = None

        self.all_callable = [method_name for method_name in dir(self) if callable(getattr(self, method_name))]

    @classmethod
    def controller_factory(cls, config=None):
        if config is None:
            config = App.instance().config.MODEL_CONTROLLER

        controller_module = config.MODULE_NAME
        controller_class = config.CLASS_NAME
        controller_module: ModelController = get_class_object_from_name(controller_module, controller_class)
        return controller_module.instance()

    def get_current_loader(self):
        return self.dataloader_controller.dataloaders[self.get_state().loader_name]

    def update_state(self, changed):
        for key in changed.keys():
            if hasattr(ModelStateController, key) and callable(getattr(ModelStateController, key)):
                self.MODEL_STATE_CONTROLLER.__getattribute__(key)(changed[key])
            else:
                self.get_state().__setattr__(key, changed[key])

    def get_state(self) -> ModelState:
        return self.MODEL_STATE_CONTROLLER.current_state

    def get_main_state(self) -> ModelState:
        return self.MODEL_STATE_CONTROLLER.states[self.MODEL_STATE_CONTROLLER.processing_stack[1][0]]

    # def reserve_state_change(self, state_name):
    #     if self.MODEL_STATE_CONTROLLER.current_state.reserve_state is not None:
    #         self.MODEL_STATE_CONTROLLER.current_state.reserve_state = state_name
    #     else:
    #         raise ValueError('double change')

    def change_state_name(self, state_name):
        self.MODEL_STATE_CONTROLLER.current_state.state_name = state_name

    def epoch_start(self):
        self.MODEL_STATE_CONTROLLER.iterator_reset()

    def step_finish(self):
        self.get_state().step += 1

    def epoch_finish(self):
        while self.get_state().state_name == "EPOCH_END":
            self.get_state().epoch += 1
            if self.get_current_loader().config.mode == 'train':
                self.OPTIMIZER.schedule()
            if self.get_state().epoch < self.get_current_loader().config.obj_epoch:
                self.change_state_name("EPOCH_START")
            else:
                self.MODEL_STATE_CONTROLLER.pop_state()

    def batch_run(self):
        self.epoch_start()
        self.get_state().tqdm.desc = "[{}({}/{})] ".format(self.get_state().loader_name, self.get_state().epoch,
                                                          self.get_current_loader().config.obj_epoch)
        self.change_state_name('BATCH_START')
        while True:
            changed = self.COMMAND_CONTROLLER.run()
            self.update_state(changed)
            state_name = self.get_state().state_name

            if state_name == 'BATCH_START': # end
                self.step_finish()
                self.change_state_name('BATCH_END')
            elif state_name == 'BATCH_END':
                self.change_state_name('BATCH_START')
            elif state_name == "EPOCH_END":
                return
            elif state_name == "START":
                return
            else:
                raise NotImplementedError

    def run(self):
        temp_bar = loading()
        while True:
            changed = self.COMMAND_CONTROLLER.run()
            self.update_state(changed)

            if self.get_state().loader_name == 'idle':
                print('\ridle' + next(temp_bar), end='')
                time.sleep(0.5)
                continue

            state_name = self.get_state().state_name
            if state_name == 'START':
                self.change_state_name('EPOCH_START')
            elif state_name == 'EPOCH_START' or state_name.split('_')[0] == 'BATCH':
                self.batch_run()
            elif state_name == 'EPOCH_END':
                self.epoch_finish()
            else:
                raise NotImplementedError

def loading():
    while True:
        for i in range(4):
            yield '.' * i



from pathlib import Path
from framework.ipc.RunningCommand import RunnableGraph
from queue import Queue

class DisplayRunnableModule(object):
    display_filename_prefix_middle = '├──'
    display_filename_prefix_last = '└──'
    display_parent_prefix_middle = '    '
    display_parent_prefix_last = '│   '

    def __init__(self, name, parent_path, is_last):
        """
        #https://stackoverflow.com/questions/9727673/list-directory-tree-structure-in-python

        :param path:
        :param parent_path:
        :param is_last:
        """
        self.name = name
        self.parent = parent_path
        self.is_last = is_last
        if self.parent:
            self.depth = self.parent.depth + 1
        else:
            self.depth = 0

    @staticmethod
    def get_ordered_module(controller, base):
        children = controller.get_runnable_module(base).required
        outs = []
        indegree = {}
        q = Queue()
        for name in children:
            indegree[name] = int(controller.get_runnable_module(name).indegree)
            if indegree[name] == 0:
                q.put(name)

        while not q.empty():
            cur_name = q.get()
            outs.append(cur_name)
            for next_name in controller.next[cur_name]:
                indegree[next_name]-=1
                if indegree[next_name]==0:
                    q.put(next_name)
        return outs

    @classmethod
    def make_tree(cls, controller, root, parent=None, is_last=False):
        displayable_root = cls(root, parent, is_last)
        yield displayable_root

        children = cls.get_ordered_module(controller, root)
        count = 1
        for path in children:
            is_last = count == len(children)
            if isinstance(controller.get_runnable_module(path), RunnableGraph):
                yield from cls.make_tree(controller, path,
                                         parent=displayable_root,
                                         is_last=is_last)
            else:
                yield cls(path, displayable_root, is_last)
            count += 1

    def display_name(self, controller):
        if isinstance(controller.get_runnable_module(self.name), RunnableGraph):
            return self.name + '/'
        return self.name

    def displayable(self, controller):
        if self.parent is None:
            return self.display_name(controller)

        _filename_prefix = (self.display_filename_prefix_last
                            if self.is_last
                            else self.display_filename_prefix_middle)

        parts = ['{!s} {!s}'.format(_filename_prefix,
                                    self.display_name(controller))]

        parent = self.parent
        while parent and parent.parent is not None:
            parts.append(self.display_parent_prefix_middle
                         if parent.is_last
                         else self.display_parent_prefix_last)
            parent = parent.parent

        return ''.join(reversed(parts))



import threading
from contextlib import contextmanager

# ref : https://gist.github.com/tylerneylon/a7ff6017b7a1f9a506cf75aa23eacfd6#file-rwlock-py
class READ_WRITE_LOCK(object):
    def __init__(self):
        self.reader = 0

        self.write_lock = threading.Lock()
        self.reader_lock = threading.Lock()

    def read_acquire(self):
        self.reader_lock.acquire()
        self.reader += 1
        if self.reader == 1:
            self.write_lock.acquire()
        self.reader_lock.release()

    def read_release(self):
        self.reader_lock.acquire()
        self.reader -= 1
        if self.reader == 0:
            self.write_lock.release()
        self.reader_lock.release()

    def write_acquire(self):
        self.write_lock.acquire()

    def write_release(self):
        self.write_lock.release()

    @contextmanager
    def r_locked(self):
        try:
            self.read_acquire()
            yield
        finally:
            self.read_release()


    @contextmanager
    def w_locked(self):
        try:
            self.write_acquire()
            yield
        finally:
            self.write_release()



import setuptools

with open("README.MD", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r", encoding="utf-8") as fh:
    requirements = fh.read().splitlines()

setuptools.setup(
    name="deep-leaps",
    version="0.0.1",
    author="Leaps",
    author_email="leap1568@gmail.com",
    description="Data driven development based deep learning framework(pytorch)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Longseabear/deep-leaps-pytorch.git",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    requirements=requirements
)


from framework.dataloader.transforms import *

"""
All transform methods are registered here. If the default transform method path is here, import the transform method.
"""
TRANSFORM = {}
TRANSFORM['ToTensor'] = ToTensor
TRANSFORM['ToNumpy'] = ToNumpy


class MainGraphFinishedException(Exception):
    def __init__(self):
        super().__init__("MainGraph Finished")
    value = property(lambda self: object(), lambda self, v: None, lambda self: None)


class MainGraphStepInterrupt(Exception):
    def __init__(self):
        super().__init__()
    value = property(lambda self: object(), lambda self, v: None, lambda self: None)


class RunnableModuleDestroyException(Exception):
    def __init__(self):
        super().__init__("This runnable module was expired.")

